-- VvardenfellTreasureChestTracker, by Gordias

VvardenfellTreasureChestTracker = {
	name = "VvardenfellTreasureChestTracker",
	version = 1.3,
	default = {
		PinData = {
		},
		texture = "esoui/art/icons/housing_bre_con_treasurechest001.dds",
		compassMaxDistance = 0.05,
		pinRotationRate = 3,
		worldPinsFlag = true,
		pinSize = 20,
		pinColor = { r = 1, g = 1, b = 1, a = 1 },
		pinDataVersion = 1.12, -- Previous version is default to force it update
		worldPinTexture = "esoui/art/icons/mapkey/mapkey_bank.dds",
		worldPinColor = { r = 1, g = 1, b = 1, a = 1 },
	}
}

local LAM = LibAddonMenu2
local LMP = LibMapPins
local L3D = Lib3D

local iconList = {
	[1] = "esoui/art/icons/housing_bre_con_treasurechest001.dds",
	[2] = "esoui/art/icons/mapkey/mapkey_bank.dds",
	[3] = "esoui/art/icons/housing_alt_fur_treasurechest001.dds",
}

local iconText = {
	[1] = "Treasure Chest  |t::esoui/art/icons/housing_bre_con_treasurechest001.dds|t",
	[2] = "Bank  |t::esoui/art/icons/mapkey/mapkey_bank.dds|t",
	[3] = "Treasure Chest II  |t::esoui/art/icons/housing_alt_fur_treasurechest001.dds|t",
}

local treasureCompass = "Vvardenfell Treasure Compass"

local treasurePin = "Vvardenfell Treasure Chests"

local pinColor = { r = 1, g = 1, b = 1, a = 1 }

local worldPinColor = { r = 1, g = 1, b = 1, a = 1 }

local compassLayout = { 
	maxDistance = 0.05,
	texture = "esoui/art/icons/housing_bre_con_treasurechest001.dds",
	sizeCallback = function(pin, angle, normalizedAngle, normalizedDistance) 
					pin:SetDimensions(32 - 32 * math.abs(normalizedDistance), 32 - 32 * math.abs(normalizedDistance))
		end,
	additionalLayout = {
		function(pin, angle, normalizedAngle, normalizedDistance) 
			if pin.pinType == treasureCompass then 
				local icon = pin:GetNamedChild("Background")
				icon:SetColor(pinColor.r, pinColor.g, pinColor.b, pinColor.a) end 
		end,
		function(pin) end
	}
}

local pinLayout = {
	level = 50,
	texture = "esoui/art/icons/housing_bre_con_treasurechest001.dds",
	size = 20,
	tint = ZO_ColorDef:New(pinColor.r,pinColor.g,pinColor.b,pinColor.a)
}

local worldPinLayout = {
	texture = "esoui/art/icons/mapkey/mapkey_bank.dds",
}

local pinTooltip = {
	creator = function(pin)
		local locX, locY = pin:GetNormalizedPosition()
		InformationTooltip:AddLine("Treasure Chest")
		end,
	tooltip = 1,
}

local panelData = {
	type = "panel",
	name = "Vvardenfell Treasure Chest Tracker",
	author = "Gordias",
	version = "1.3",
	registerForRefresh = true,
	registerForDefaults = true,
}

-- Compass callback function is called everytime, the compass is refreshed.
local function CompassCallback(pinManager)

	-- Return if the map is world, alliance or cosmic
	if GetMapType() > MAPTYPE_ZONE then return end

	-- Return if the zone is NOT Vvardenfell
	local zoneName = GetUnitZone("player")
	if not string.match(zoneName, "Vvardenfell") then return end

	local mapname = LMP:GetZoneAndSubzone(true)
	local pins = VvardenfellTreasureChestTracker.savedVariables.PinData[mapname]

	-- Return if there is no pin at the current map
	if not pins then return end

	for _, pinInfo in ipairs(pins) do
		COMPASS_PINS.pinManager:CreatePin(treasureCompass, pinInfo, pinInfo.x, pinInfo.y)
	end

end

-- VvardenfellTreasureChestTracker:PinCallback(pinManager) will be called every time map is changed.
local function PinCallback(pinManager)

	-- Return if the pinType is not enabled
	if not LMP:IsEnabled(treasurePin) then return end

	-- Return if the map is world, alliance or cosmic
	if GetMapType() > MAPTYPE_ZONE then return end

	local mapname = LMP:GetZoneAndSubzone(true)
	local pins = VvardenfellTreasureChestTracker.savedVariables.PinData[mapname]

	-- Return if there is no pin at the current map
	if not pins then return end

	for _, pinInfo in ipairs(pins) do
		LMP:CreatePin(treasurePin, pinInfo, pinInfo.x, pinInfo.y)
	end

end

-- VvardenfellTreasureChestTracker.RegisterVisualControls(event) will be called on EVENT_PLAYER_ACTIVATED
local function RegisterVisualControls()

	-- Return if 3D Pins are NOT enabled
	if not VvardenfellTreasureChestTracker.savedVariables.worldPinsFlag then return end

	-- Return if the zone is NOT Vvardenfell
	local zoneName = GetUnitZone("player")
	if not string.match(zoneName, "Vvardenfell") then return end

	-- Create the 3D mode for parent
	VTCTWindow:Create3DRenderSpace()

	-- Create the control and 3D mode for pins
	for i = 1, 3 do
		local worldPinControl = VTCTWindow:GetChild(i)
		worldPinControl:Create3DRenderSpace()
		worldPinControl:Set3DLocalDimensions(1, 1)

		-- Set the texture
		worldPinControl:SetTexture(worldPinLayout.texture)

		-- Set the world pin Color
		local worldPinColor = VvardenfellTreasureChestTracker.savedVariables.worldPinColor
		worldPinControl:SetColor(worldPinColor.r, worldPinColor.g, worldPinColor.b, worldPinColor.a)

		-- Set the world pin Blend Mode (whatever the heaven that is)
		worldPinControl:SetBlendMode(TEX_BLEND_MODE_ALPHA)
	end

end

-- VvardenfellTreasureChestTracker.OnBeginLockpick(event) will be called every time a lockpicking event occurs
local function OnBeginLockpick(event)

	-- Return if interactable is NOT a chest
	-- TODO : Add "Safebox" (and interactableName ~= "Safebox" ) to the list if you can verify they spawn inside them as well
	local _, interactableName, _, _, _, _, _, _ = GetGameCameraInteractableActionInfo()
	if interactableName ~= "Chest" then return end

	-- Return if the zone is NOT Vvardenfell
	local zoneName = GetUnitZone("player")
	if not string.match(zoneName, "Vvardenfell") then return end

	-- Return if the map is NOT "vvardenfell/vvardenfell_base"
	local mapname = LMP:GetZoneAndSubzone(true)
	if mapname ~=  "vvardenfell/vvardenfell_base" then return end

	-- Take the current position and height of the player
	local posX, posY, posHeading = GetMapPlayerPosition("player")
	local _, height, _ = VTCTWindow:Get3DRenderSpaceOrigin()

	-- Create the array if it has not been already
	VvardenfellTreasureChestTracker.savedVariables.PinData[mapname] = VvardenfellTreasureChestTracker.savedVariables.PinData[mapname] or {}
	
	local pins = VvardenfellTreasureChestTracker.savedVariables.PinData[mapname]

	local pinEpsilon = 0.001

	-- Check if the pin has already been created before (check the neighborhood with pinEpsilon)
	for i, pinInfo in ipairs(pins) do
		if math.abs(pinInfo.x - posX) < pinEpsilon and math.abs(pinInfo.y - posY) < pinEpsilon then 
			-- Found a match, check if the height data is inserted
			if pinInfo.h == nil then
				table.remove(VvardenfellTreasureChestTracker.savedVariables.PinData[mapname], i)
				table.insert(VvardenfellTreasureChestTracker.savedVariables.PinData[mapname], i, {x = pinInfo.x, y = pinInfo.y, h = height})
			end
			-- The pin is already saved, return
			return 
		end
	end

	-- Add the newPinInfo to the PinData
	local newPinInfo = { x = posX, y = posY, h = height }
	table.insert(VvardenfellTreasureChestTracker.savedVariables.PinData[mapname], newPinInfo)

	-- Refresh Compass and Pins
	LMP:RefreshPins(treasurePin)
	COMPASS_PINS:RefreshPins(treasureCompass)

end

local function OnUpdateVisuals()

	-- Return if 3D Pins are NOT enabled
	if not VvardenfellTreasureChestTracker.savedVariables.worldPinsFlag then return end

	-- Return if the zone is NOT Vvardenfell
	local zoneName = GetUnitZone("player")
	if not string.match(zoneName, "Vvardenfell") then return end

	-- Get mapname
	local mapname = LMP:GetZoneAndSubzone(true)

	-- Load the pins
	local pins = VvardenfellTreasureChestTracker.savedVariables.PinData[mapname]

	-- Return if there is no pin at the current map
	if not pins then return end

	-- Take player's current position and search for the pin closest to the player
	-- TODO : Use quicksort and take top 3 = faster
	local playerX, playerY = GetMapPlayerPosition("player")
	local dist = 2
	local closestPins = { 
		[1] = { x = playerX, y = playerY, h = 120.0, d = 2 },
		[2] = { x = playerX, y = playerY, h = 120.0, d = 2 },
		[3] = { x = playerX, y = playerY, h = 120.0, d = 2 },
	}

	for _, pinInfo in ipairs(pins) do
		dist = math.sqrt(math.pow(pinInfo.x - playerX, 2) + math.pow(pinInfo.y - playerY, 2))
		local valH
		if  pinInfo.h == nil then
			valH = 120.0
		else
			valH = pinInfo.h
		end
		if closestPins[3].d > dist then
			if closestPins[2].d > dist then
				if closestPins[1].d > dist then
					closestPins[3] = closestPins[2]
					closestPins[2] = closestPins[1]
					closestPins[1] = {x = pinInfo.x, y = pinInfo.y, h = valH, d = dist}
				else
					closestPins[3] = closestPins[2]
					closestPins[2] = {x = pinInfo.x, y = pinInfo.y, h = valH, d = dist}
				end
			else
			closestPins[3] = {x = pinInfo.x, y = pinInfo.y, h = valH, d = dist}
			end
		end
	end


	for i, closePins in pairs(closestPins) do 
		
		-- Pin coordinates are in local, conver them to World (L3D) and then to GUI (ZOS World is in cm while L3D World is in cm)
		local worldX, worldZ = L3D:LocalToWorld(closePins.x, closePins.y)
		local renderX, renderY, renderZ = WorldPositionToGuiRender3DPosition(worldX * 100.0, closePins.h * 100.0, worldZ * 100.0)

		-- Take current pin control from the parent
		local worldPinControl = VTCTWindow:GetChild(i)

		-- Set the pin position
		worldPinControl:Set3DRenderSpaceOrigin(renderX, renderY, renderZ)

		-- Set the pin dimensions
		local dimensionScale = 1
		worldPinControl:Set3DLocalDimensions(dimensionScale, dimensionScale)

		-- Set the pin orientation
		if VvardenfellTreasureChestTracker.savedVariables.pinRotationRate == 0 then
			-- If no rotation is selected, face the pin texture to the player
			local camHeading = GetPlayerCameraHeading()
			worldPinControl:Set3DRenderSpaceOrientation(0, camHeading, 0)
		else
			local _, p, _ = worldPinControl:Get3DRenderSpaceOrientation()
			worldPinControl:Set3DRenderSpaceOrientation(0, p + math.rad(VvardenfellTreasureChestTracker.savedVariables.pinRotationRate), 0)
		end

	end

	-- Check if the player is lockpicking - hide the first (closest) control if that is the case
	local InteractionType = GetInteractionType()
	local worldPinControl = VTCTWindow:GetChild(1)
    if (InteractionType == INTERACTION_LOCKPICK) then 
		worldPinControl:SetHidden(true)
	else
		worldPinControl:SetHidden(false)
	end

end

-- Addon optionsData for LibAddonMenu-2.0 (LAM)
local optionsData = {
	[1] = {
		type = "header",
		name = "General Settings",
	},
	[2] = {
		type = "dropdown",
		name = "Pin Icon",
		tooltip = "Sets the pin icon",
		choices = iconText,
		getFunc = function() 
			local tmpText = VvardenfellTreasureChestTracker.savedVariables.texture
			if tmpText == iconList[1] then
				return iconText[1] 
			elseif tmpText == iconList[2] then 
				return iconText[2]
			elseif tmpText == iconList[3] then 
				return iconText[3]
			end
		end,
		setFunc = function(var)
			local tmpText = iconList[1]
			if var == iconText[1] then
				tmpText = iconList[1]
			elseif var == iconText[2] then
				tmpText = iconList[2]
			elseif var == iconText[3] then
				tmpText = iconList[3]
			end
			VvardenfellTreasureChestTracker.savedVariables.texture = tmpText
			pinLayout.texture = tmpText
			LMP:SetLayoutData(treasurePin, pinLayout)
			LMP:RefreshPins(treasurePin)
			compassLayout.texture = tmpText
			COMPASS_PINS:RefreshPins(treasureCompass)

		end,
		default = iconText[1],
	},
	[3] = {
		type = "slider",
		name = "Pin Size",
		tooltip = "Pin will be resized to this value",
		min = 16,
		max = 32,
		step = 4,
		getFunc = function() return VvardenfellTreasureChestTracker.savedVariables.pinSize end,
		setFunc = function(value) VvardenfellTreasureChestTracker.savedVariables.pinSize = value 
				pinLayout.size = value end,
		default = 20,
	},
	[4] = {
		type = "slider",
		name = "Compass Distance",
		tooltip = "The maximum distance for the pin to be visible",
		min = 5,
		max = 200,
		step = 1,
		getFunc = function() return (VvardenfellTreasureChestTracker.savedVariables.compassMaxDistance * 1000) end,
		setFunc = function(value) VvardenfellTreasureChestTracker.savedVariables.compassMaxDistance = (value / 1000)
				compassLayout.maxDistance = (value / 1000) end,
		default = 50,
	},
	[5] = {
		type = "colorpicker",
		name = "Pin Color",
		tooltip = "Set Pin Color",
		getFunc = function() return pinColor.r, pinColor.g, pinColor.b, pinColor.a end,
		setFunc = function(r,g,b,a) pinColor.r = r pinColor.g = g pinColor.b = b pinColor.a = a 
				VvardenfellTreasureChestTracker.savedVariables.pinColor = pinColor
				pinLayout.tint = ZO_ColorDef:New(pinColor.r, pinColor.g, pinColor.b, pinColor.a)
				LMP:SetLayoutData(treasurePin, pinLayout)
				LMP:RefreshPins(treasurePin)
			end,
		default = ZO_ColorDef:New(1,1,1,1),
	},
	[6] = {
		type = "header",
		name = "3D Pins (Beta)",
	},
	[7] = {
		type = "checkbox",
		name = "Enable 3D Pins",
		tooltip = "Enable/Disable 3D Pins",
		getFunc = function() return VvardenfellTreasureChestTracker.savedVariables.worldPinsFlag end,
		setFunc = function(value) 
				if value then 
					VvardenfellTreasureChestTracker.savedVariables.worldPinsFlag = true
					RegisterVisualControls()
					EVENT_MANAGER:RegisterForUpdate(VvardenfellTreasureChestTracker.name,  30, OnUpdateVisuals)
				else
					VvardenfellTreasureChestTracker.savedVariables.worldPinsFlag = false
					VTCTWindow:Destroy3DRenderSpace()
					EVENT_MANAGER:UnregisterForUpdate(VvardenfellTreasureChestTracker.name, OnUpdateVisuals )
				end
			end,
		default = true,
	},
	[8] = {
		type = "slider",
		name = "Pin Rotation Rate",
		tooltip = "The rotation rate of the pins",
		min = 0,
		max = 10,
		step = 1,
		getFunc = function() return VvardenfellTreasureChestTracker.savedVariables.pinRotationRate end,
		setFunc = function(value) VvardenfellTreasureChestTracker.savedVariables.pinRotationRate = value end,
		default = 3,
	},
	[9] = {
		type = "dropdown",
		name = "World Pin Icon",
		tooltip = "Set the world pin icon",
		choices = iconText,
		getFunc = function() 
			local tmpText = VvardenfellTreasureChestTracker.savedVariables.worldPinTexture
			if tmpText == iconList[1] then
				return iconText[1] 
			elseif tmpText == iconList[2] then 
				return iconText[2]
			elseif tmpText == iconList[3] then 
				return iconText[3]
			end
		end,
		setFunc = function(var)
			local tmpText = iconList[1]
			if var == iconText[1] then
				tmpText = iconList[1]
			elseif var == iconText[2] then
				tmpText = iconList[2]
			elseif var == iconText[3] then
				tmpText = iconList[3]
			end
			VvardenfellTreasureChestTracker.savedVariables.worldPinTexture = tmpText
			worldPinLayout.texture = tmpText
			-- Update worldPinControls
			for i = 1, 3 do
				local worldPinControl = VTCTWindow:GetChild(i)
				worldPinControl:SetTexture(tmpText)
			end

		end,
		default = iconText[2],
	},
	[10] = {
		type = "colorpicker",
		name = "World Pin Color",
		tooltip = "Sets the world pin color",
		getFunc = function() return worldPinColor.r, worldPinColor.g, worldPinColor.b, worldPinColor.a end,
		setFunc = function(r,g,b,a) worldPinColor.r = r worldPinColor.g = g worldPinColor.b = b worldPinColor.a = a 
				VvardenfellTreasureChestTracker.savedVariables.worldPinColor = worldPinColor
				worldPinColor = ZO_ColorDef:New(worldPinColor.r, worldPinColor.g, worldPinColor.b, worldPinColor.a)
				-- Update worldPinControls
				for i = 1, 3 do
					local worldPinControl = VTCTWindow:GetChild(i)
					worldPinControl:SetColor(worldPinColor.r, worldPinColor.g, worldPinColor.b, worldPinColor.a)
				end
			end,
		default = ZO_ColorDef:New(1,1,1,1),
	},
}

-- Initialization function
function VvardenfellTreasureChestTracker:Initialize()

	-- Load saved variables
	VvardenfellTreasureChestTracker.savedVariables = ZO_SavedVars:New("VvardenfellTreasureChestTrackerSavedVars", 1, nil, VvardenfellTreasureChestTracker.default)

	-- Assign saved variables
	pinColor = VvardenfellTreasureChestTracker.savedVariables.pinColor
	worldPinColor = VvardenfellTreasureChestTracker.savedVariables.worldPinColor
	compassLayout.maxDistance = VvardenfellTreasureChestTracker.savedVariables.compassMaxDistance
	compassLayout.texture = VvardenfellTreasureChestTracker.savedVariables.texture
	pinLayout.tint = ZO_ColorDef:New(pinColor.r,pinColor.g,pinColor.b,pinColor.a)
	pinLayout.size = VvardenfellTreasureChestTracker.savedVariables.pinSize
	pinLayout.texture = VvardenfellTreasureChestTracker.savedVariables.texture
	worldPinLayout.texture = VvardenfellTreasureChestTracker.savedVariables.worldPinTexture

	-- Check the pinDataVersion to see if new pinData integration is necessary
	if VvardenfellTreasureChestTracker.savedVariables.pinDataVersion < VvardenfellTreasureChestTracker.version then
		local mapname = "vvardenfell/vvardenfell_base"
		-- Create the array if it has not been already
		VvardenfellTreasureChestTracker.savedVariables.PinData[mapname] = VvardenfellTreasureChestTracker.savedVariables.PinData[mapname] or {}
		local currentPinData = VvardenfellTreasureChestTrackerData_GetKnownData()
		local knownPins = currentPinData[mapname]
		local pins = VvardenfellTreasureChestTracker.savedVariables.PinData[mapname]
		local pinEpsilon = 0.001
		local foundFlag = false
		-- Check if the pin has already been found by the player (check the neighborhood with pinEpsilon)
		for _, knownPinInfo in ipairs(knownPins) do
			for i, pinInfo in ipairs(pins) do
				if math.abs(knownPinInfo.x - pinInfo.x) < pinEpsilon and math.abs(knownPinInfo.y - pinInfo.y) < pinEpsilon then
					foundFlag = true
					-- Insert the height data if it is not there already
					if pinInfo.h == nil then 
						-- First remove the line that doesn't have the height data
						table.remove(VvardenfellTreasureChestTracker.savedVariables.PinData[mapname], i)
						-- Then add the line again with the height data
						table.insert(VvardenfellTreasureChestTracker.savedVariables.PinData[mapname], i, {x = pinInfo.x, y = pinInfo.y, h = knownPinInfo.h})
					end
					break 
				end
			end
			if foundFlag == false then
				table.insert(VvardenfellTreasureChestTracker.savedVariables.PinData[mapname], {x = knownPinInfo.x, y = knownPinInfo.y, h = knownPinInfo.h})
			end
			foundFlag = false
		end
		-- Update the pinDataVersion
		VvardenfellTreasureChestTracker.savedVariables.pinDataVersion = VvardenfellTreasureChestTracker.version
	end

	-- Initialize settings
	LAM:RegisterAddonPanel("VvardenfellTreasureChestTrackerOptions", panelData)
	LAM:RegisterOptionControls("VvardenfellTreasureChestTrackerOptions", optionsData)

	-- Initialize compass pins
	COMPASS_PINS:AddCustomPin(treasureCompass , CompassCallback, compassLayout)
	COMPASS_PINS:RefreshPins(treasureCompass)

	-- Initialize map pins
	LMP:AddPinType(treasurePin, PinCallback, nil, pinLayout, pinTooltip)
	LMP:AddPinFilter(treasurePin, nil, nil, nil)
	LMP:RefreshPins(treasurePin)

	-- Register Lib3D events
	L3D:RegisterWorldChangeCallback(VvardenfellTreasureChestTracker.name, function(identifier, zoneIndex, currentZoneId, wasReload)
		local zoneName = GetUnitZone("player")
		if string.match(zoneName, "Vvardenfell") then 
			RegisterVisualControls()
			EVENT_MANAGER:RegisterForUpdate(VvardenfellTreasureChestTracker.name, 30, OnUpdateVisuals )
		else
			EVENT_MANAGER:UnregisterForUpdate(VvardenfellTreasureChestTracker.name, OnUpdateVisuals )
		end
	end)

	-- Register the tracker event : OnBeginLockpick
	EVENT_MANAGER:RegisterForEvent(VvardenfellTreasureChestTracker.name, EVENT_BEGIN_LOCKPICK, OnBeginLockpick)
	-- Unregister the initialization event (is this really necessary?)
	EVENT_MANAGER:UnregisterForEvent(VvardenfellTreasureChestTracker.name, EVENT_ADD_ON_LOADED)

end

-- Event handler function for the "addon loaded" event
local function OnAddOnLoaded(event, addonName)
	if addonName == VvardenfellTreasureChestTracker.name then
		VvardenfellTreasureChestTracker:Initialize()
	end
end
 
-- Register the initialization event
EVENT_MANAGER:RegisterForEvent(VvardenfellTreasureChestTracker.name, EVENT_ADD_ON_LOADED, OnAddOnLoaded)